# GRL Estimation
# Link Flow Assignment

rm(list=ls(all=TRUE))
setwd("D:/夏の学校2019/経路選択RL")
input_area <- "D:/夏の学校2019/経路選択RL/input"
inputLinkPass <- paste(input_area, "network/link_assignment.csv", sep = "/")
inputMatrixPass <- paste(input_area, "network/matrix_assignment.csv", sep = "/")
inputDistancePass <- paste(input_area, "network/distance_assignment.csv", sep = "/")
inputDemandPass <- paste(input_area, "demand_od_free.csv", sep = "/") # demand_od***.csv

readdata.number <- -1 # -1 = all

############### time measure ############
message("PROGRAM START!\n")
t<-proc.time()


#########################################
########  reading & sampling    #########
#########################################

######### Demand data reading ###########
data <- read.csv(inputDemandPass, nrows=readdata.number)
names(data) <- c("o", "d", "demand")
ddata <- sort(unique(data$d))
odata <- sort(unique(data$o))

########## link data reading ############
link <- read.csv(inputLinkPass, colClasses=c("integer", "integer", "integer", "double", "integer", "integer"))
linkid <- sort(unique(link$LinkID))

########## link matrix reading ##########
matrix <- read.csv(inputMatrixPass)
names(matrix) <- c("k", "a")
asetdataList <- list()
matrixList <- list()

################# index #################
D <- length(unique(data$d))             # Length of unique OD pair
L <- length(linkid)                     # No. of links
AList <- list()

for(k in 1:L){
  kl <- linkid[k]
  sub <- subset(matrix, matrix$k == kl)
  AList[[k]] <- sub$a                           # sets of oytgoing links from given link k is seperated in list
}

######### link connectiuon dummy ########
I <- array(0, dim = c(L, L))
names(matrix) <- c("k", "a")

for(i in 1:nrow(matrix)){
  kl <- matrix$k[i]
  al <- matrix$a[i]
  k <- (1:L)[linkid == kl]
  a <- (1:L)[linkid == al]
  I[k,a] <- 1                                   # Link connection incidence matrix
}

Ilist <- list()

for(d in 1:D){                                  # for d in no no. of destinations
  dl <- ddata[d]
  dn <- (1:L)[linkid == dl]
  Id <- I
  Id[dn, ] <- 0
  
  for(k in 1:L){
    aList <- AList[[k]]
    if(is.element(dl, aList)){
      Id[k, ] <- 0
      Id[k, dn] <- 1
    }
  }
  
  Ilist[[d]] <- Id
}

######### explanatory variable ##########
# リンク長
cc <- array(exp(0), dim = c(1, L))

for(a in 1:L){
  cc[1, a] <- exp(link$Length[a] / 100)
}

CC <- array(rep(cc), dim = c(L, L))
CC <- t(CC)

# 店舗数
ss <- array(exp(0), dim = c(1, L))

for(a in 1:L){
  ss[1, a] <- exp(link$N_shops[a] / 50)
}

SS <- array(rep(ss), dim = c(L, L))
SS <- t(SS)

# 駐車場からの距離
Distance <- read.csv(inputDistancePass)
ad <- nrow(Distance) 
dd <- ncol(Distance)
DList <- list()

for(d in 1:D){
  dn <- (1:L)[linkid == ddata[d]]
  tt <- array(exp(0), dim = c(1, L))
  for(a in 1:ad){
    tt[1, a] <- exp(Distance[a, dn] / 200)
  }
  TT <- array(rep(tt), dim = c(L, L))
  DList[[d]] <- t(TT)
}


#########################################
############ value function #############
#########################################

z <- matrix(1, nrow = L, ncol = D) #exp(Vd)

# parameter
b <- c(-0.1704, 0.3098, -0.6062) # リンク長、店舗数、駐車場距離
beta <- 0.00000002 # 時間割引率
plist <- list()

for (d in 1:D){
  kd <- (1:length(linkid))[linkid == ddata[d]] #####d
  z[kd, d] <- 1                                #####d
  
  M <- matrix(0, nrow = L, ncol = L) #
  B <- matrix(0, nrow = L, ncol = 1) #
  B[kd, 1] <- 1                      #destination link is 1
  Id <- Ilist[[d]] #d              
  TT <- DList[[d]]
  
  # 瞬間効用Mを計算
  for (k in 1:L){
    for (a in 1:L){
      Ika <- Id[k, a]
      if(Ika == 1){
        if(a == kd){
          M[k, a] <- 1
        }else{
          M[k, a] <- (CC[k, a] ^ b[1]) * (SS[k, a] ^ b[2]) * (TT[k, a] ^ b[3])
        }
      }
    }
  } 
  
  # 価値関数ベクトルzの求解
  #ID <- diag(L)
  #VF <- solve(ID - M) %*% B
  #VF <- VF ^ beta
  # ↑は誤り
  
  dL <- 100
  i <- 0
  zi <- matrix(z[,d], nrow = L, ncol = 1)
  
  while(dL > 0.01){
    zii <- zi
    Xz <- (zii) ^ beta
    zi <- M %*% Xz + B
    dL <- sum(abs(zii - zi))
    i <- i + 1
  }
  
  z[,d] <- zi
  
  # リンク選択確率を計算
  PLM <- matrix(0, nrow=L, ncol=L) # Link probability matrix
  PL <- list()                    # list for PLM
  
  for(n in 1:nrow(M)){
    if(M[n, ] %*% z[, d] > 0){             # to avoid infinity value in calculation
      tr <- M[n, ] * t(z[, d])
      do <- matrix(rep((M[n, ] %*% z[, d])), nrow=1, ncol=L)
      Prob <- tr / do
    }
    PL[[n]] <- Prob
    PLM[n, ] <- PL[[n]]
    
    for(r in 1:D){
      kg <- (1:length(linkid))[linkid == ddata[r]]            # set 0 probabability for destination to destination
      PLM[kg, ] <- 0 
    }
    
    k <- ddata[[d]]
    
    for (m in 1:nrow(M)){
      p <- AList[[m]]
      
      if(is.element(k,p)){
        PLM[m, kd] <- 1                                       # set 0 probability of extra links to destination
      }else{
        PLM[m, kd] <- 0
      }
    }
  }
  
  plist[[d]] <- PLM
}

print(plist)

######### Link Flow Estimation ############
flist <- list()
id <- diag(L)

for (b in 1:D){
  # (I - P) の逆行列
  mt <- as.matrix(plist[[b]])
  In <- (id - t(mt))
  so <- solve(In)
  
  # 発生交通量G
  v <- matrix(0, nrow=L, ncol=1)
  
  for (c in 1:nrow(data)) {
    t <- data$d[c]
    if(ddata[b] == t){
      n <- data$o[c]
      y <- (1:length(linkid))[linkid == n]
      v[y, 1] <- data$demand[c]
    }
  }
  
  flow <- so %*% v
  flist[[b]] <- flow
}

# sum of link flows for multiple destinations
fv <- matrix(0, nrow= L, ncol=1)

for (g in 1:length(flist)){
  fv <- fv + flist[[g]]
  rownames(fv) <- as.character(linkid)
}

options("scipen"=1000, "digits"=0)
print(fv)


########## calculation time #############
message("ALL COMPLETED!\n")
print(proc.time()-t)

write.csv(fv, "link_flow_free.csv")
